from html import escape as _escape_html
import logging
import smtplib
import socket
import textwrap

from flask_mail import Message
from flask import current_app

import keg_mail.utils as utils

__all__ = [
    'Email',
    'EmailContent',
    'SendEmailError',
    'escape_html',
    'send_email',
]


log = logging.getLogger(__name__)


def escape_html(content):
    """Takes any string and converts it to an `EmailContent` object where the HTML variant has
    been escaped."""
    return EmailContent(html=_escape_html(content), text=content)


def send_email(recipient, email):
    """Sends an `Email` to a recipient.

    :param recipient: is an email address.
    :param email: is an `Email` object.

    :raises: `SendEmailError`.
    """

    mail_engine = current_app.extensions.get('mail', None)

    if not mail_engine:  # pragma: no cover
        raise SendEmailError('No mail engine')

    message = Message(
        email.subject,
        recipients=[recipient],
        html=email.content.html,
        body=email.content.text
    )

    try:
        mail_engine.send(message)
    except (socket.gaierror, socket.error):  # pragma: no cover
        raise SendEmailError('SMTP Connection error')
    except smtplib.SMTPAuthenticationError:  # pragma: no cover
        raise SendEmailError('SMTP Authentication error')

    if current_app.config.get('MAIL_LOG_MESSAGES', False):
        log.info(message)


class SendEmailError(Exception):
    """Failure to send an email."""
    pass


class EmailContent(object):
    """Captures any content designed to go in an email, including both plaintext and HTML variants.

    This class allows both plaintext and HTML variants of email content to be combined into a single
    entity with a common interface.
    """

    def __init__(self, html, text, clean_whitespace=True):
        """Captures both HTML and text variants of the content.
        :param clean_whitespace: if True content will be dedented and stripped of extra whitespace.
        """
        self.html = textwrap.dedent(html).strip() if clean_whitespace else html
        self.text = textwrap.dedent(text).strip() if clean_whitespace else text

    def format(self, *args, **kwargs):
        """Just like str.format but applies the formatting to both the plaintext and HTML variants.

        If any of the format arguments are `EmailContent` objects, then the appropriate variant will
        be extracted and applied to the corresponding variant in the template.
        """
        select_html = lambda item: getattr(item, 'html', item)  # noqa
        select_text = lambda item: getattr(item, 'text', item)  # noqa
        return EmailContent(
            html=self.html.format(*map(select_html, args), **utils.dict_map(select_html, kwargs)),
            text=self.text.format(*map(select_text, args), **utils.dict_map(select_text, kwargs))
        )

    def join(self, contents):
        """Intercalates this email content between an iterable of other email contents."""
        eval_contents = list(contents)
        return EmailContent(
            html=self.html.join(x.html for x in eval_contents),
            text=self.text.join(x.text for x in eval_contents)
        )

    def __eq__(self, other):
        try:
            return self.html == other.html and self.text == other.text
        except AttributeError:  # pragma: no cover
            return False

    def __hash__(self):
        return hash((self.html, self.text))


class Email(object):
    """Captures the full content of an email: subject and body.

    This class is very much like `EmailContent` but additionally includes a subject.
    """

    def __init__(self, subject, content, type_=None):
        """
        :param subject: the email subject as a string. It may be a template.
        :param content: is a string or an `EmailContent` object. It may be a template.
        """
        self.subject = textwrap.dedent(subject).strip()
        self.content = content
        self.type_ = type_

    def format(self, *args, **kwargs):
        select_text = lambda item: getattr(item, 'text', item)  # noqa
        return Email(
            self.subject.format(*map(select_text, args),
                                **utils.dict_map(select_text, kwargs)),
            self.content.format(*args, **kwargs),
            type_=self.type_
        )

    def __eq__(self, other):
        try:
            return (self.subject == other.subject
                    and self.content == other.content
                    and self.type_ == other.type_)
        except AttributeError:  # pragma: no cover
            return False

    def __hash__(self):
        return hash((self.subject, self.content, self.type_))



import keg_mail.content as content


class TestEmailContent(object):
    def test_content_is_dedented(self):
        text = """
        stuff
        thing
        """
        cont = content.EmailContent(text, text)

        assert cont.text == "stuff\nthing"
        assert cont.html == "stuff\nthing"

    def test_content_isnt_dedented(self):
        text = """
        stuff
        thing
        """
        cont = content.EmailContent(text, text, clean_whitespace=False)
        assert cont.text == text
        assert cont.html == text

    def test_format_applys_kwargs(self):
        text = '{var}'

        cont = content.EmailContent(text, text)
        cont_replaced = cont.format(var='thing')
        assert cont != cont_replaced
        assert cont_replaced.text == 'thing'
        assert cont_replaced.html == 'thing'

    def test_join_combines_contents(self):
        cont1 = content.EmailContent('====', '====')
        cont2 = content.EmailContent('stuff', 'stuff')
        cont3 = content.EmailContent('thing', 'thing')
        cont4 = cont1.join([cont2, cont3])

        assert cont4.text == 'stuff====thing'
        assert cont4.html == 'stuff====thing'

    def test_equality(self):
        assert content.EmailContent('a', 'a') == content.EmailContent('a', 'a')
        assert content.EmailContent('b', 'a') != content.EmailContent('a', 'a')
        assert content.EmailContent('a', 'b') != content.EmailContent('a', 'a')
        assert content.EmailContent('b', 'b') != content.EmailContent('a', 'a')


class TestEmail(object):

    def setup_method(self):
        self.content = content.EmailContent('{a}', '{a}')
        self.content_b = content.EmailContent('{b}', '{b}')

    def test_equality(self):
        # Same
        assert (content.Email('a', self.content, 'a') ==
                content.Email('a', self.content, 'a'))

        # Different subject
        assert (content.Email('a', self.content, 'a') !=
                content.Email('b', self.content, 'a'))

        # Different content
        assert (content.Email('a', self.content, 'a') !=
                content.Email('a', self.content_b, 'a'))

        # Different type
        assert (content.Email('a', self.content, 'a') !=
                content.Email('a', self.content, 'b'))



import hashlib
import hmac

import flask
import keg
from keg.db import db
from keg.web import BaseView


__all__ = [
    'WebhookBase',
    'LogStatusWebhook',
]


class WebhookBase(BaseView):
    def verify(self, signing_key, data):
        signature = data['signature']

        hmac_digest = hmac.new(
            key=signing_key.encode(),
            msg='{}{}'.format(signature['timestamp'], signature['token']).encode(),
            digestmod=hashlib.sha256
        ).hexdigest()
        return hmac.compare_digest(str(signature['signature']), str(hmac_digest))

    def post(self):
        data = flask.request.json

        signing_key = keg.current_app.config['MAIL_MAILGUN_WEBHOOK_KEY']
        if not self.verify(signing_key, data):
            flask.abort(403)

        self.process_event(data['event-data'])

        return 'OK', 200

    def process_event(self, event_data):
        pass


class LogStatusWebhook(WebhookBase):
    def process_event(self, event_data):
        try:
            keg.current_app.extensions['mail'].mailgun_update_message_status(event_data)
        except Exception:
            # Ensure any row locks are released
            db.session.rollback()
            raise



import cgi
import io
import sys
import urllib.parse
from datetime import timedelta
from decimal import Decimal
from operator import attrgetter
from unittest import mock

import arrow
import flask_mail
import keg_mail
from keg.db import db

from keg_mail.mailgun import MailgunClient

import keg_mail_ta.model.entities as ents


class TestMailgunClient:
    def setup_method(self, _):
        ents.MailLog.delete_cascaded()

    def test_send(self, requests_mock):
        requests_mock.post('https://api.mailgun.net/v3/example.com/messages',
                           json={'response': 'ok'})

        client = MailgunClient(domain='example.com', api_key='foo', testing=False)
        message = flask_mail.Message(
            subject='RE: xyz',
            recipients=['bob@example.com', 'joe@example.com'],
            body='very important message',
            html='<blink>very important message</blink>',
            sender='no-reply@aol.com',
            cc=['tim@example.com'],
            bcc=['al@example.com'],
        )
        resp = client.send(message)

        assert resp == {'response': 'ok'}

        assert requests_mock.call_count == 1

        req = requests_mock.request_history[0]
        body = urllib.parse.parse_qs(req.text)

        assert body['from'] == ['no-reply@aol.com']
        assert body['to'] == ['bob@example.com', 'joe@example.com']
        assert body['cc'] == ['tim@example.com']
        assert body['bcc'] == ['al@example.com']
        assert body['subject'] == ['RE: xyz']
        assert body['text'] == ['very important message']
        assert body['html'] == ['<blink>very important message</blink>']
        assert 'attachment' not in body

    def test_send_attachments(self, requests_mock):
        requests_mock.post('https://api.mailgun.net/v3/example.com/messages',
                           json={'response': 'ok'})

        client = MailgunClient(domain='example.com', api_key='foo', testing=False)
        message = flask_mail.Message(
            subject='RE: xyz',
            recipients=['bob@example.com', 'joe@example.com'],
            body='very important message',
            html='<blink>very important message</blink>',
            sender='no-reply@aol.com',
            cc=['tim@example.com'],
            bcc=['al@example.com'],
            attachments=[
                flask_mail.Attachment(
                    filename='foo.jpg',
                    content_type='image/jpeg',
                    data=b'image data',
                ),
                flask_mail.Attachment(
                    filename='bar.pdf',
                    content_type='application/pdf',
                    data=b'document data',
                ),
            ]
        )
        resp = client.send(message)

        assert resp == {'response': 'ok'}

        assert requests_mock.call_count == 1

        req = requests_mock.request_history[0]

        ctype, pdict = cgi.parse_header(req.headers['Content-Type'])
        pdict['boundary'] = pdict['boundary'].encode()
        pdict['CONTENT-LENGTH'] = str(len(req.body))
        body = cgi.parse_multipart(io.BytesIO(req.body), pdict)

        def t(val):
            # Prior to python 3.7 cgi parse results were bytes
            return val.encode() if sys.version_info[:2] < (3, 7) else val

        assert body['from'] == [t('no-reply@aol.com')]
        assert body['to'] == [t('bob@example.com'), t('joe@example.com')]
        assert body['cc'] == [t('tim@example.com')]
        assert body['bcc'] == [t('al@example.com')]
        assert body['subject'] == [t('RE: xyz')]
        assert body['text'] == [t('very important message')]
        assert body['html'] == [t('<blink>very important message</blink>')]
        assert body['attachment'] == [b'image data', b'document data']

    def test_sync_webhooks(self, requests_mock):
        requests_mock.get('https://api.mailgun.net/v3/domains/example.com/webhooks', json={
            "webhooks": {
                "delivered": {
                    "urls": [
                        "https://example.com/delivered_hook1",
                        "https://example.com/delivered_hook2"
                    ]
                },
                "permanent_fail": {
                    "urls": ["https://example.com/perm_fail_hook"]
                },
                "opened": {
                    "urls": ["https://example.com/opened_hook"]
                }
            }
        })
        requests_mock.delete('https://api.mailgun.net/v3/domains/example.com/webhooks/delivered',
                             json={'response': 'ok'})
        requests_mock.post('https://api.mailgun.net/v3/domains/example.com/webhooks',
                           json={'response': 'ok'})
        requests_mock.put('https://api.mailgun.net/v3/domains/example.com/webhooks/permanent_fail',
                          json={'response': 'ok'})

        webhook_def = {
            "permanent_fail": [
                "perm_fail_hook1",
                "perm_fail_hook2",
            ],
            "temporary_fail": ["temp_fail_hook"],
            "opened": ["opened_hook"]
        }

        with mock.patch('flask.url_for') as m_url_for:
            m_url_for.side_effect = lambda x, **kwargs: 'https://example.com/{}'.format(x)
            client = MailgunClient(domain='example.com', api_key='foo', testing=False)
            client.sync_webhooks(webhook_def)

        assert requests_mock.call_count == 4

        history = sorted(requests_mock.request_history, key=attrgetter('method'))

        req1 = history[0]
        assert req1.method == 'DELETE'

        req2 = history[1]
        assert req2.method == 'GET'

        req3 = history[2]
        assert req3.method == 'POST'
        req3_body = urllib.parse.parse_qs(req3.text)
        assert req3_body['id'] == ['temporary_fail']
        assert req3_body['url'] == ['https://example.com/temp_fail_hook']

        req4 = history[3]
        assert req4.method == 'PUT'
        req4_body = urllib.parse.parse_qs(req4.text)
        assert req4_body['url'] == [
            'https://example.com/perm_fail_hook1',
            'https://example.com/perm_fail_hook2',
        ]

    def test_clear_webhooks(self, requests_mock):
        requests_mock.get('https://api.mailgun.net/v3/domains/example.com/webhooks', json={
            "webhooks": {
                "delivered": {
                    "urls": [
                        "https://example.com/delivered_hook1",
                        "https://example.com/delivered_hook2"
                    ]
                },
                "permanent_fail": {
                    "urls": ["https://example.com/perm_fail_hook"]
                },
            }
        })
        requests_mock.delete('https://api.mailgun.net/v3/domains/example.com/webhooks/delivered',
                             json={'response': 'ok'})
        requests_mock.delete(
            'https://api.mailgun.net/v3/domains/example.com/webhooks/permanent_fail',
            json={'response': 'ok'}
        )

        client = MailgunClient(domain='example.com', api_key='foo', testing=False)
        client.clear_webhooks()

        assert requests_mock.call_count == 3

        history = sorted(requests_mock.request_history, key=attrgetter('url'))

        assert history[0].method == 'GET'
        assert history[0].url == 'https://api.mailgun.net/v3/domains/example.com/webhooks'

        assert history[1].method == 'DELETE'
        assert history[1].url == 'https://api.mailgun.net/v3/domains/example.com/webhooks/delivered'

        assert history[2].method == 'DELETE'
        assert history[2].url == (
            'https://api.mailgun.net/v3/domains/example.com/webhooks/permanent_fail'
        )

    def test_poll_events(self, requests_mock):
        now = arrow.utcnow().float_timestamp

        log_ent = ents.MailLog.testing_create(
            status=keg_mail.EmailLogStatus.sent,
            status_updated=arrow.utcnow().shift(hours=-2)
        )

        requests_mock.get('https://api.mailgun.net/v3/example.com/events', json={
            "items": [
                {
                    "id": "a" * 22,
                    "timestamp": now - 60 * 60,
                    "event": "failed",
                    "severity": "temporary",
                    "message": {
                        "headers": {
                          "to": "foo@example.com",
                          "message-id": "20130812164300.28108.52546@samples.mailgun.org",
                          "from": "Excited User <me@samples.mailgun.org>",
                          "subject": "Hello"
                        },
                        "attachments": [],
                        "recipients": [
                            "foo@example.com",
                            "baz@example.com",
                            "bar@example.com"
                        ],
                        "size": 80
                    },
                    "recipient": "baz@example.com",
                    "method": "http"
                }
            ],
            'paging': {
                'next': 'https://api.mailgun.net/v3/example.com/events/page2'
            }
        })

        requests_mock.get('https://api.mailgun.net/v3/example.com/events/page2', json={
            "items": [
                {
                    "id": "b" * 22,
                    "timestamp": now - 25 * 60,
                    "event": "accepted",
                    "message": {
                        "headers": {
                            "to": "foo@example.com",
                            "message-id": "20130812164300.28108.52546@samples.mailgun.org",
                            "from": "Excited User <me@samples.mailgun.org>",
                            "subject": "Hello"
                        },
                        "attachments": [],
                        "recipients": [
                            "foo@example.com",
                            "baz@example.com",
                            "bar@example.com"
                        ],
                        "size": 81
                    },
                    "recipient": "baz@example.com",
                    "method": "http"
                }
            ],
        })

        client = MailgunClient(
            domain='example.com', api_key='foo', testing=False,
            log_entity_cls=ents.MailLog
        )
        with mock.patch.object(client, 'update_message_status') as m_status:
            client.poll_events()

        history = requests_mock.request_history
        assert len(history) == 2

        assert history[0].path == '/v3/example.com/events'
        assert history[0].qs['ascending'] == ['yes']
        assert history[0].qs['begin'] == [str(log_ent.status_updated.float_timestamp)]
        assert arrow.get(float(history[0].qs['end'][0])).is_between(
            arrow.utcnow().shift(minutes=-1),
            arrow.utcnow()
        )
        assert history[0].qs['limit'] == ['300']

        assert history[1].url == 'https://api.mailgun.net/v3/example.com/events/page2'

        m_status.assert_called_once()
        args, kwargs = m_status.call_args
        assert len(args) == 1
        assert args[0]['id'] == 'a' * 22
        assert kwargs == {'_commit': False}

        # Both pages trustworthy
        with mock.patch.object(client, 'update_message_status') as m_status:
            client.poll_events(threshold=timedelta(minutes=10))

        assert m_status.call_count == 2
        assert len(history) == 4
        assert history[2].path == '/v3/example.com/events'
        assert history[3].url == 'https://api.mailgun.net/v3/example.com/events/page2'

        # Neither page is trustworthy
        with mock.patch.object(client, 'update_message_status') as m_status:
            client.poll_events(threshold=timedelta(minutes=120))

        assert m_status.call_count == 0
        assert len(history) == 5
        assert history[4].path == '/v3/example.com/events'

    def test_update_message_status(self, requests_mock):
        now = arrow.utcnow()
        log_ent = ents.MailLog.testing_create(
            status=keg_mail.EmailLogStatus.sent,
            status_updated=now.shift(hours=-2),
            message_id='foo',
            address='baz@example.com'
        )

        event_data = {
            "id": "a" * 22,
            "timestamp": now.float_timestamp,
            "event": "failed",
            "severity": "temporary",
            "message": {
                "headers": {
                    "to": "foo@example.com",
                    "message-id": "foo",
                    "from": "Excited User <me@samples.mailgun.org>",
                    "subject": "Hello"
                },
                "attachments": [],
                "recipients": [
                    "foo@example.com",
                    "baz@example.com",
                    "bar@example.com"
                ],
                "size": 80
            },
            'delivery-status': {
                'message': 'Temp Error'
            },
            "recipient": "baz@example.com",
            "method": "http"
        }

        client = MailgunClient(
            domain='example.com', api_key='foo', testing=False,
            log_entity_cls=ents.MailLog
        )
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.failed_temporary
        assert log_ent.status_updated == now

        # event timestamp is older than current status
        event_data['timestamp'] = now.shift(seconds=-1).float_timestamp
        event_data['severity'] = 'permanent'
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.failed_temporary
        assert log_ent.status_updated == now
        assert log_ent.error_detail == 'Temp Error'

        # ignored event type
        event_data['event'] = 'opened'
        event_data['timestamp'] = log_ent.status_updated.shift(seconds=1).float_timestamp
        del event_data['severity']
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.failed_temporary
        assert log_ent.status_updated == now
        assert log_ent.error_detail == 'Temp Error'

        # record not found
        event_data['event'] = 'delivered'
        event_data['message']['headers']['message-id'] = 'bar'
        event_data['timestamp'] = log_ent.status_updated.shift(seconds=1).float_timestamp
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.failed_temporary
        assert log_ent.status_updated == now
        assert log_ent.error_detail == 'Temp Error'

        # non error status
        event_data['event'] = 'delivered'
        event_data['message']['headers']['message-id'] = 'foo'
        event_data['timestamp'] = log_ent.status_updated.shift(seconds=1).float_timestamp
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.delivered
        assert log_ent.status_updated == now.shift(seconds=1)
        assert log_ent.error_detail is None

        # rejected status
        event_data['event'] = 'rejected'
        event_data['reject'] = {'reason': 'Some Reason'}
        event_data['timestamp'] = log_ent.status_updated.shift(seconds=1).float_timestamp
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.rejected
        assert log_ent.status_updated == now.shift(seconds=2)
        assert log_ent.error_detail == 'Some Reason'

        # accepts decimal for timestamp
        event_data['timestamp'] = Decimal.from_float(
            log_ent.status_updated.shift(seconds=1).float_timestamp
        )
        client.update_message_status(event_data)
        db.session.expire(log_ent)
        assert log_ent.status == keg_mail.EmailLogStatus.rejected
        assert log_ent.status_updated == now.shift(seconds=3)
        assert log_ent.error_detail == 'Some Reason'



import enum
import typing

import sqlalchemy as sa
from sqlalchemy_utils import (
    EmailType,
    ArrowType,
)


__all__ = [
    'EmailLogStatus',
    'MailLogMixin',
]


class EmailLogStatus(enum.Enum):
    sent = 'Sent'
    rejected = 'Rejected'
    delivered = 'Delivered'
    failed_permanent = 'Failed (Permanent)'
    failed_temporary = 'Failed (Temporary)'

    @classmethod
    def db_name(cls):
        return 'enum_email_logs_status'

    @classmethod
    def from_mailgun_event(cls, event: str, severity: typing.Optional[str]):
        mapping = {
            'accepted': cls.sent,
            'rejected': cls.rejected,
            'delivered': cls.delivered,
            'failed': cls.failed_permanent if severity == 'permanent' else cls.failed_temporary
        }
        return mapping.get(event)

    @classmethod
    def error_statuses(cls):
        return [
            cls.rejected,
            cls.failed_temporary,
            cls.failed_permanent,
        ]

    @property
    def is_error(self):
        return self in self.error_statuses()


class MailLogMixin:
    address = sa.Column(EmailType(), nullable=False)
    subject = sa.Column(sa.Unicode, nullable=False)
    status = sa.Column(sa.Enum(EmailLogStatus, name='enum_email_logs_status'), nullable=False)
    status_updated = sa.Column(ArrowType, nullable=True)
    message_id = sa.Column(sa.Unicode)
    message_uuid = sa.Column(sa.Unicode)
    error_detail = sa.Column(sa.Unicode)

    # __table_args__ = (
    #     sa.UniqueConstraint(address, message_id),
    #     sa.UniqueConstraint(message_uuid)
    # )



from keg.db import db
from keg_elements.db.mixins import DefaultColsMixin, MethodsMixin

from keg_mail.model import MailLogMixin


class MailLog(db.Model, DefaultColsMixin, MethodsMixin, MailLogMixin):
    __tablename__ = 'mail_log'



import json
import logging
import uuid

import flask_mail


__all__ = [
    'KegMail',
]

from werkzeug.utils import import_string

from keg_mail import cli
from keg_mail.model import EmailLogStatus

log = logging.getLogger(__name__)


class _MailMixin(flask_mail._MailMixin):
    def mailgun_send(self, message):
        return self.mailgun_client.send(message)

    def mailgun_sync_webhooks(self):
        return self.mailgun_client.sync_webhooks(self.mailgun_webhooks)

    def mailgun_clear_webhooks(self):
        return self.mailgun_client.clear_webhooks()

    def mailgun_poll_events(self):
        return self.mailgun_client.poll_events()

    def log_message_sent(self, message, mailgun_response=None, message_uuid=None, app=None,
                         _commit=True):
        if not self.enable_logging:
            return

        from keg.db import db

        message_id = mailgun_response.get('id') if mailgun_response else None

        for recipient in message.recipients:
            entry = self.log_entity_cls(
                address=recipient,
                subject=message.subject,
                status=EmailLogStatus.sent,
                message_id=message_id,
                message_uuid=message_uuid,
            )
            db.session.add(entry)
        db.session.commit()

    def mailgun_update_message_status(self, event_data):
        if not self.enable_logging:
            return
        self.mailgun_client.update_message_status(event_data)

    def send(self, message):
        if self.mailgun_client:
            return self.mailgun_send(message)

        if message.extra_headers is None:
            message.extra_headers = {}
        message.extra_headers['X-Mailgun-Variables'] = json.dumps(
            {'message_uuid': str(uuid.uuid4())}
        )
        return super().send(message)


class _Mail(_MailMixin):
    def __init__(self, server, username, password, port, use_tls, use_ssl,
                 default_sender, debug, max_emails, suppress, mailgun_api_key, mailgun_domain,
                 mailgun_webhooks, enable_logging, log_entity_cls, ascii_attachments=False):
        self.server = server
        self.username = username
        self.password = password
        self.port = port
        self.use_tls = use_tls
        self.use_ssl = use_ssl
        self.default_sender = default_sender
        self.debug = debug
        self.max_emails = max_emails
        self.suppress = suppress
        self.ascii_attachments = ascii_attachments

        self.mailgun_api_key = mailgun_api_key
        self.mailgun_domain = mailgun_domain
        self.mailgun_webhooks = mailgun_webhooks
        self.enable_logging = enable_logging
        self.log_entity_cls = log_entity_cls


class KegMail(flask_mail.Mail):
    def init_mail(self, config, debug=False, testing=False):
        mailgun_domain = config.get('MAIL_MAILGUN_DOMAIN')
        mailgun_api_key = config.get('MAIL_MAILGUN_API_KEY')
        mailgun_test_mode = config.get('MAIL_MAILGUN_TEST_MODE', False)
        enable_logging = config.get('MAIL_ENABLE_LOGGING', False)
        log_entity_loc = config.get('MAIL_LOG_ENTITY')
        log_entity_cls = import_string(log_entity_loc) if log_entity_loc else None
        state = _Mail(
            server=config.get('MAIL_SERVER', '127.0.0.1'),
            username=config.get('MAIL_USERNAME'),
            password=config.get('MAIL_PASSWORD'),
            port=config.get('MAIL_PORT', 25),
            use_tls=config.get('MAIL_USE_TLS', False),
            use_ssl=config.get('MAIL_USE_SSL', False),
            default_sender=config.get('MAIL_DEFAULT_SENDER'),
            debug=int(config.get('MAIL_DEBUG', debug)),
            max_emails=config.get('MAIL_MAX_EMAILS'),
            suppress=config.get('MAIL_SUPPRESS_SEND', testing),
            ascii_attachments=config.get('MAIL_ASCII_ATTACHMENTS', False),
            mailgun_api_key=mailgun_api_key,
            mailgun_domain=mailgun_domain,
            mailgun_webhooks=config.get('MAIL_MAILGUN_WEBHOOKS', {}),
            enable_logging=enable_logging,
            log_entity_cls=log_entity_cls,
        )

        if enable_logging and not log_entity_cls:
            raise ValueError('When mail logging is enabled you must specify a logging entity class')

        if enable_logging:
            flask_mail.email_dispatched.connect(state.log_message_sent)

        state.mailgun_client = None
        if mailgun_domain and mailgun_api_key:
            from keg_mail import mailgun
            state.mailgun_client = mailgun.MailgunClient(
                mailgun_domain,
                mailgun_api_key,
                mailgun_test_mode,
                log_entity_cls=log_entity_cls
            )

        return state

    def init_app(self, app):
        state = super().init_app(app)
        self.init_cli(app, cli_group_name=app.config.get('MAIL_CLI_GROUP_NAME', 'mail'))
        return state

    def init_cli(self, app, cli_group_name='mail'):
        cli.add_cli_to_app(app, cli_group_name=cli_group_name)

    def send(self, message):
        # Flask Mail's inheritance pattern for this class is weird. No actual instance of this class
        # should exist following initialization. Instead an instance of the _Mail class above gets
        # installed as the extension.
        raise NotImplementedError



from keg.app import Keg

from keg_mail.plugin import KegMail

from keg_mail_ta.views import km_blueprint


class KegMailTestApp(Keg):
    import_name = 'keg_mail_ta'
    db_enabled = True
    use_blueprints = [km_blueprint]
    keyring_enable = False

    def on_init_complete(self):
        self.mail = KegMail(self)
        return self


if __name__ == '__main__':
    from keg_mail_ta import app
    app.KegMailTestApp.cli.main()



import hashlib
import hmac
from unittest import mock

import arrow
import flask
import flask_webtest
from keg_mail import views


@mock.patch.object(views.WebhookBase, 'process_event')
class TestWebhookBase:
    def test_good_signature(self, m_process):
        ts = str(int(arrow.utcnow().float_timestamp)).encode()
        token = b'b' * 32
        key = b'a' * 36

        body = {
            'signature': {
                'timestamp': ts.decode(),
                'token': token.decode(),
                'signature': hmac.new(key=key, msg=ts + token, digestmod=hashlib.sha256).hexdigest()
            },
            'event-data': {'foo': 'bar'}
        }

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.post_json('/noop-webhook', body)

        assert resp.status_code == 200
        assert resp.text == 'OK'

        m_process.assert_called_once_with({'foo': 'bar'})

    def test_bad_signature(self, m_process):
        ts = str(int(arrow.utcnow().float_timestamp)).encode()
        token = b'b' * 32
        key = b'a' * 35 + b'b'

        body = {
            'signature': {
                'timestamp': ts.decode(),
                'token': token.decode(),
                'signature': hmac.new(key=key, msg=ts + token, digestmod=hashlib.sha256).hexdigest()
            },
            'event-data': {}
        }

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.post_json('/noop-webhook', body, status=403)

        assert resp.status_code == 403
        assert not m_process.called


@mock.patch('flask.current_app.extensions', autospec=True, spec_set=True)
class TestLogStatusWebhook:
    def test_calls_update_status_method(self, m_extensions):
        ts = str(int(arrow.utcnow().float_timestamp)).encode()
        token = b'b' * 32
        key = b'a' * 36

        body = {
            'signature': {
                'timestamp': ts.decode(),
                'token': token.decode(),
                'signature': hmac.new(key=key, msg=ts + token, digestmod=hashlib.sha256).hexdigest()
            },
            'event-data': {'foo': 'bar'}
        }

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.post_json('/log-status-webhook', body)

        assert resp.status_code == 200
        assert resp.text == 'OK'

        m_update = m_extensions.__getitem__.return_value.mailgun_update_message_status
        m_update.assert_called_once_with({'foo': 'bar'})



import os.path as osp
from setuptools import setup, find_packages


cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'README.rst')).read()
CHANGELOG = open(osp.join(cdir, 'changelog.rst')).read()

version = {}
with open(osp.join(cdir, 'keg_mail', 'version.py')) as version_fp:
    exec(version_fp.read(), version)

setup(
    name="Keg-Mail",
    description="A mail sending library for keg applications",
    version=version['VERSION'],
    long_description='\n\n'.join((README, CHANGELOG)),
    author="Level 12 Developers",
    author_email="devteam@level12.io",
    url='https://github.com/level12/keg-mail',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    license='BSD',
    packages=find_packages(),
    zip_safe=True,
    install_requires=[
        'flask_mail',
        'Keg',
        'SQLAlchemy-Utils',
        'arrow',
    ],
    extras_require={
        'mailgun': [
            'requests',
            'ordered-set',
        ],
        'test': [
            'tox',
            'pytest',
            'pytest-coverage',
            'requests-mock',
            'flask-webtest',
            'KegElements',
        ]
    }
)



from blazeutils.strings import randchars


class DefaultProfile(object):
    SECRET_KEY = randchars()

    # These three just get rid of warnings on the console.
    KEG_KEYRING_ENABLE = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    SITE_NAME = 'Keg Mail Demo'
    SITE_ABBR = 'KM Demo'


class TestProfile(object):
    MAIL_MAILGUN_WEBHOOK_KEY = 'a' * 36



import keg


def add_cli_to_app(app, cli_group_name='mail'):
    mail_ext = app.extensions['mail']

    if mail_ext.mailgun_domain is None or mail_ext.mailgun_api_key is None:
        return

    @app.cli.group(cli_group_name)
    def mail():
        pass

    @mail.command('sync-webhooks')
    def _sync_mailgun_webhooks():
        keg.current_app.extensions['mail'].mailgun_sync_webhooks()

    @mail.command('remove-webhooks')
    def _remove_mailgun_webhooks():
        keg.current_app.extensions['mail'].mailgun_clear_webhooks()

    @mail.command('poll-events')
    def _poll_events():
        keg.current_app.extensions['mail'].mailgun_poll_events()

    mail_ext.cli_group = mail



import flask

from keg_mail.views import (
    LogStatusWebhook as LogStatusWebhookBase,
    WebhookBase,
)

km_blueprint = flask.Blueprint('keg_mail', __name__)


class NoOpWebhook(WebhookBase):
    blueprint = km_blueprint
    url = '/noop-webhook'


class LogStatusWebhook(LogStatusWebhookBase):
    blueprint = km_blueprint
    url = '/log-status-webhook'



import json
import logging
import uuid
from datetime import timedelta
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import arrow
import flask
import flask_mail
import requests
from blazeutils.helpers import ensure_list
from flask import current_app
from ordered_set import OrderedSet

from keg_mail.model import (
    EmailLogStatus,
    MailLogMixin,
)

log = logging.getLogger(__name__)


class MailgunError(Exception):
    pass


FileData = Union[BinaryIO, bytes]

# Covers the data layouts that requests accepts while allowing multiple files with the same
# multi-part field name
RequestFile = Tuple[
    str,  # field name
    Union[
        FileData,
        # filename, file obj
        Tuple[str, FileData],
        # filename, file obj, content type
        Tuple[str, FileData, str],
        # filename, file obj, content type, custom headers
        Tuple[str, FileData, str, Dict[str, Any]]
    ]
]


class MailgunClient:
    api_root_url = 'https://api.mailgun.net/v3'

    def __init__(self, domain: str, api_key: str, testing: bool,
                 log_entity_cls: Optional[Type[MailLogMixin]] = None):
        self.domain = domain
        self.api_key = api_key
        self.log_entity_cls = log_entity_cls
        self.testing = testing

    def _get_auth(self):
        return requests.auth.HTTPBasicAuth('api', self.api_key)

    def _endpoint_url(self, endpoint: str):
        return '{}/{}/{}'.format(self.api_root_url, self.domain, endpoint.lstrip('/'))

    def _request(self, endpoint: str, method: Callable,
                 body: Optional[Dict] = None,
                 params: Optional[Dict] = None,
                 files: Optional[List[RequestFile]] = None,
                 allowed_status=200):

        url = endpoint if endpoint.startswith('http') else self._endpoint_url(endpoint)
        auth = self._get_auth()

        resp = method(url, data=body, params=params, auth=auth, files=files)
        if resp.status_code != allowed_status:
            raise MailgunError(resp.text)

        try:
            resp_data = resp.json()
        except json.JSONDecodeError as err:
            raise MailgunError(str(err))

        return resp_data

    def _post_request(self, endpoint: str, body: Dict,
                      params: Optional[Dict] = None,
                      files: Optional[List[RequestFile]] = None, allowed_status=200):
        return self._request(endpoint, requests.post, body, params, files, allowed_status)

    def _get_request(self, endpoint: str, query_data: Optional[Dict] = None,
                     allowed_status=200):
        return self._request(endpoint, requests.get, None, query_data,
                             allowed_status=allowed_status)

    def _delete_request(self, endpoint: str, query_data: Optional[Dict] = None,
                        allowed_status=200):
        return self._request(endpoint, requests.delete, None, query_data,
                             allowed_status=allowed_status)

    def _put_request(self, endpoint: str, body: Dict,
                     params: Optional[Dict] = None,
                     files: Optional[List[RequestFile]] = None, allowed_status=200):
        return self._request(endpoint, requests.put, body, params, files, allowed_status)

    def _ensure_unique_recipients(self, to: List[str], cc: List[str], bcc: List[str]):
        # We count on the recipients being unique per message_id so the same address in both the
        # "to" and "cc" fields is a problem.
        to = OrderedSet(to)  # Remove duplicates from "to" addresses
        cc = OrderedSet(cc) - to  # If an address is in "to" it needn't be in "cc"
        bcc = OrderedSet(bcc) - to - cc  # If an address is in "to" or "cc" it needn't be in "bcc"

        return list(to), list(cc), list(bcc)

    def send(self, message: flask_mail.Message):
        to, cc, bcc = self._ensure_unique_recipients(
            ensure_list(message.recipients),
            ensure_list(message.cc),
            ensure_list(message.bcc)
        )
        message_uuid = uuid.uuid4()
        request_body = {
            'from': message.sender,
            'to': to,
            'cc': cc,
            'bcc': bcc,
            'subject': message.subject,
            'text': message.body,
            'html': message.html,
            'v:message_uuid': message_uuid,
            'o:testmode': 'yes' if self.testing else 'no'
        }

        files = None
        if message.attachments:
            files = [
                ('attachment', (attachment.filename, attachment.data, attachment.content_type))
                for attachment in message.attachments
            ]

        resp = self._post_request('messages', request_body, files=files)
        flask_mail.email_dispatched.send(
            message,
            mailgun_response=resp,
            message_uuid=message_uuid,
            app=current_app._get_current_object(),
        )
        return resp

    def poll_events(self, after: Optional[arrow.Arrow] = None,
                    threshold=timedelta(minutes=30)):
        from keg.db import db
        import sqlalchemy as sa

        if not self.log_entity_cls:
            raise ValueError('KegMail must be configured with MAIL_LOG_ENTITY to use this feature')

        if after is None:
            after = db.session.query(
                sa.func.max(self.log_entity_cls.status_updated)
            ).filter(
                self.log_entity_cls.status_updated.isnot(None)
            ).scalar()
            after = after if after else arrow.utcnow().shift(minutes=-30)

        for event in self._iter_events(after, threshold=threshold):
            try:
                self.update_message_status(event, _commit=False)
            except Exception:
                # Make sure we release any locked records
                db.session.rollback()
                raise
        db.session.commit()

    def _iter_events(self, after: arrow.Arrow, threshold=timedelta(minutes=30)):
        # Pages of events newer than 30 minutes ago should not be considered complete and
        # trustworthy according to the API documentation:
        # https://documentation.mailgun.com/en/latest/api-events.html#event-polling

        now = arrow.utcnow()

        json_resp = self._get_request('events', {
            'ascending': 'yes',
            'begin': after.float_timestamp,
            'end': now.float_timestamp,
            'limit': 300  # API defined max
        })

        while True:
            if not json_resp['items']:
                break

            #  check if the page is trustworthy
            newest_ts = arrow.get(json_resp['items'][-1]['timestamp'])
            if newest_ts + threshold > now:
                return

            for event in json_resp['items']:
                yield event

            # Get url for next page of results
            next_url = json_resp['paging'].get('next') if json_resp.get('paging') else None
            if next_url is None:
                break

            # Load the next page
            resp = requests.get(next_url, auth=self._get_auth())
            json_resp = resp.json()
            if resp.status_code != 200:
                raise MailgunError(json_resp)

    def update_message_status(self, event_data, _commit=True):
        from keg.db import db

        if not self.log_entity_cls:
            raise ValueError('KegMail must be configured with MAIL_LOG_ENTITY to use this feature')

        new_status = EmailLogStatus.from_mailgun_event(
            event_data['event'],
            event_data.get('severity')
        )
        if new_status is None:
            return

        recipient = event_data['recipient']
        if event_data.get('user-variables'):
            message_uuid = event_data['user-variables'].get('message_uuid')
            # Lock the record in case it is currently being updated via webhook
            log_entry = self.log_entity_cls.query.filter_by(
                message_uuid=message_uuid,
                address=recipient,
            ).with_for_update().first()
        else:
            message_id = event_data['message']['headers']['message-id']
            # Lock the record in case it is currently being updated via webhook
            log_entry = self.log_entity_cls.query.filter_by(
                address=recipient,
                message_id=message_id
            ).with_for_update().first()

        if log_entry is None:
            if _commit:
                db.session.rollback()  # release row lock
            return

        timestamp = arrow.get(float(event_data['timestamp']))
        if log_entry.status_updated and log_entry.status_updated >= timestamp:
            if _commit:
                db.session.rollback()  # release row lock
            return

        log_entry.status = new_status
        log_entry.status_updated = timestamp
        if new_status == EmailLogStatus.rejected:
            log_entry.error_detail = event_data['reject']['reason']
        elif new_status.is_error:
            log_entry.error_detail = event_data['delivery-status']['message']
        else:
            log_entry.error_detail = None

        if _commit:
            db.session.commit()

    def sync_webhooks(self, webhook_defs: dict):
        def url(path: str = ''):
            return '{}/domains/{}/webhooks/{}'.format(
                self.api_root_url, self.domain, path.lstrip('/')).rstrip('/')

        resp = self._get_request(url())
        installed_hooks = resp['webhooks']

        hooks = {}
        for event, endpoints in webhook_defs.items():
            hooks[event] = []
            for endpoint in endpoints:
                if not isinstance(endpoint, str):
                    endpoint, args = endpoint
                else:
                    args = {}
                hooks[event].append(
                    flask.url_for(endpoint, **args, _external=True, _scheme='https')
                )

        # remove installed hooks that are no longer used
        for event in installed_hooks:
            if event not in webhook_defs:
                log.info('Removing {} hooks'.format(event))
                self._delete_request(url(event))

        # update installed webhooks
        for event, urls in hooks.items():
            if event not in installed_hooks:
                log.info('[Keg Mail] Adding {} hooks: {}'.format(event, urls))
                self._post_request(url(), {'id': event, 'url': urls})
            elif installed_hooks[event]['urls'] != urls:
                log.info('Updating {} hooks: {} -> {}'.format(
                    event, installed_hooks[event]['urls'], urls))
                self._put_request(url(event), {'url': urls})

    def clear_webhooks(self):
        resp = self._get_request('{}/domains/{}/webhooks'.format(self.api_root_url, self.domain))
        installed_hooks = resp['webhooks']

        for event in installed_hooks:
            log.info('Removing {} hooks'.format(event))
            url = '{}/domains/{}/webhooks/{}'.format(self.api_root_url, self.domain, event)
            self._delete_request(url)
